import pandas as pd
from SmileGAN.Smile_GAN_clustering import cross_validated_clustering, single_model_clustering
import os

if __name__ == '__main__':
	output_dir = './smile'
	if not os.path.exists(output_dir):
		os.makedirs(output_dir)

	train_data = pd.read_csv('sample_data.csv')

	ncluster = 3
	start_saving_epoch = 6000
	max_epoch = 14000
	WD = 0.14
	AQ = 60
	cluster_loss = 0.002


	cross_validated_clustering(train_data, ncluster, 4, 0.8, start_saving_epoch, max_epoch, output_dir, WD, AQ, cluster_loss,\
		'highest_matching_clustering', batchSize=60, lipschitz_k=0.5)






import os
import torch
import torch.nn.functional as F
from collections import OrderedDict
from torch.autograd import Variable
from itertools import chain as ichain
from .networks import define_Linear_Mapping, define_Linear_Clustering, define_Linear_Discriminator

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"


#####sample from discrete uniform random variable and construct SUB variable. 
def sample_z(real_X, ncluster, fix_class=-1):
    """
    sample from discrete uniform random variable and construct SUB variable.

    :param real_X, torch tensor with real data 
    :param ncluster: int, defined number of clusters
    :param fix_class: int, can be set to certain mapping directions to generate data in only one cluster,
                           set to -1 for random sampling from discrete uniform distribution

    :return tensor with sampled or selected mapping directions & 
            tensors with shape n*k (each row is a one-hot vector depending on sampled directions) 
    """
    Tensor = torch.FloatTensor
    z = Tensor(real_X.size(0), ncluster).fill_(0)
    z_idx = torch.empty(real_X.size(0), dtype=torch.long)
    if (fix_class == -1):
        z_idx = z_idx.random_(ncluster)
        z = z.scatter_(1, z_idx.unsqueeze(1), 1.)
    else:
        z_idx[:] = fix_class
        z = z.scatter_(1, z_idx.unsqueeze(1), 1.)
        z[0,fix_class]
    z_var = Variable(z)
    return z_var, z_idx

def criterion_GAN(pred, target_is_real):
    if target_is_real:
        target_var = Variable(pred.data.new(pred.shape[0]).long().fill_(0.))
        loss=F.cross_entropy(pred, target_var)
    else:
        target_var = Variable(pred.data.new(pred.shape[0]).long().fill_(1.))
        loss = F.cross_entropy(pred, target_var)
    return loss

class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

class SmileGAN(object):
    def __init__(self):
        self.opt = None

        ##### definition of all netwotks
        self.netMapping = None
        self.netClustering= None
        self.netDiscriminator = None

        ##### definition of all optimizers
        self.optimizer_M = None
        self.optimizer_D = None

        ##### definition of all criterions
        self.criterionGAN = criterion_GAN
        self.criterionChange = F.l1_loss
        self.criterionCluster=F.cross_entropy


    def create(self, opt):

        self.opt = opt

        ## definition of all netwotks
        self.netMapping = define_Linear_Mapping(self.opt.nROI,self.opt.ncluster)
        self.netClustering = define_Linear_Clustering(self.opt.nROI,self.opt.ncluster)
        self.netDiscriminator = define_Linear_Discriminator(self.opt.nROI,self.opt.ncluster)

        ## definition of all optimizers
        self.optimizer_M = torch.optim.Adam(ichain(self.netMapping.parameters(),self.netClustering.parameters()),
                                        lr=self.opt.lr, betas=(self.opt.beta1, 0.999),weight_decay=2.5*1e-3)
        self.optimizer_D = torch.optim.Adam(ichain(self.netDiscriminator.parameters()),
                                            lr=self.opt.lr/5., betas=(self.opt.beta1, 0.999))


    def train_instance(self, x, real_y):
        z_var, z_index = sample_z(x, self.opt.ncluster)
        fake_y = self.netMapping.forward(x,z_var)+x

        ## Discriminator loss
        pred_fake_y = self.netDiscriminator.forward(fake_y.detach())
        loss_D_fake_y = self.criterionGAN(pred_fake_y, False)
        pred_true_y = self.netDiscriminator.forward(real_y)
        loss_D_true_y = self.criterionGAN(pred_true_y, True)
        loss_D= 0.5* (loss_D_fake_y + loss_D_true_y)

        ## update weights of discriminator
        self.optimizer_D.zero_grad()
        loss_D.backward()
        gnorm_D = torch.nn.utils.clip_grad_norm_(self.netDiscriminator.parameters(), self.opt.max_gnorm)
        self.optimizer_D.step()

        ## Mapping/Change/Cluster loss
        pred_fake_y = self.netDiscriminator.forward(fake_y)
        loss_mapping = self.criterionGAN(pred_fake_y, True)

        reconst_z_softmax,reconst_z = self.netClustering.forward(fake_y)
        cluster_loss = self.criterionCluster(reconst_z, z_index)

        change_loss= self.criterionChange(fake_y, x)
        loss_G = loss_mapping+self.opt.lam*cluster_loss+self.opt.mu*change_loss

        ## update weights of mapping and clustering function
        self.optimizer_M.zero_grad()
        loss_G.backward()
        gnorm_M = torch.nn.utils.clip_grad_norm_(self.netMapping.parameters(), self.opt.max_gnorm)
        self.optimizer_M.step()

        ## perform weight clipping
        for p in self.netMapping.parameters():
            p.data.clamp_(-self.opt.lipschitz_k, self.opt.lipschitz_k)
        for p in self.netClustering.parameters():
            p.data.clamp_(-self.opt.lipschitz_k, self.opt.lipschitz_k)

        ## Return dicts
        losses=OrderedDict([('Discriminator_loss', loss_D.item()),('Mapping_loss', loss_mapping.item()),('loss_change', change_loss.item()),('loss_cluster', cluster_loss.item())])

        return losses
    
    ## return a numpy array of pattern type probabilities for all input subjects
    def predict_cluster(self,real_y):
        prediction,z=self.netClustering.forward(real_y)
        return prediction.detach().numpy()

    ## return generated patient data with given zub variable
    def predict_Y(self, x, z):
        return self.netMapping.forward(x, z)+x

    ## save checkpoint    
    def save(self, save_dir, chk_name):
        chk_path = os.path.join(save_dir, chk_name)
        checkpoint = {
            'netMapping':self.netMapping.state_dict(),
            'netDiscriminator':self.netDiscriminator.state_dict(),
            'optimizer_D':self.optimizer_D.state_dict(),
            'optimizer_M':self.optimizer_M.state_dict(),
            'netClustering':self.netClustering.state_dict(),
        }
        checkpoint.update(self.opt)
        torch.save(checkpoint, chk_path)

    def load_opt(self,checkpoint):
        self.opt = dotdict({})
        for key in checkpoint.keys():
            if key not in ['netMapping','netDiscriminator','netClustering','optimizer_M','optimizer_D']:
                self.opt[key] = checkpoint[key]
        

    ## load trained model
    def load(self, chk_path):
        checkpoint = torch.load(chk_path)
        self.load_opt(checkpoint)
        ##### definition of all netwotks
        self.netMapping = define_Linear_Mapping(self.opt.nROI,self.opt.ncluster)
        self.netClustering = define_Linear_Clustering(self.opt.nROI,self.opt.ncluster)
        self.netDiscriminator = define_Linear_Discriminator(self.opt.nROI,self.opt.ncluster)

        ##### definition of all optimizers
        self.optimizer_M = torch.optim.Adam(ichain(self.netMapping.parameters(),self.netClustering.parameters()),
                                        lr=self.opt.lr, betas=(self.opt.beta1, 0.999),weight_decay=2.5*1e-3)
        self.optimizer_D = torch.optim.Adam(ichain(self.netDiscriminator.parameters()),
                                        lr=self.opt.lr/5., betas=(self.opt.beta1, 0.999))
        
        self.netMapping.load_state_dict(checkpoint['netMapping'])
        self.netDiscriminator.load_state_dict(checkpoint['netDiscriminator'])
        self.netClustering.load_state_dict(checkpoint['netClustering'])
        self.optimizer_D.load_state_dict(checkpoint['optimizer_D'])
        self.optimizer_M.load_state_dict(checkpoint['optimizer_M'])
        self.load_opt(checkpoint)
            


        



import os
import time
import numpy as np
from tqdm import tqdm
from torch.autograd import Variable
from .model import SmileGAN
from .evaluate import eval_w_distances, label_change
from .utils import Covariate_correction, Data_normalization, parse_train_data, parse_validation_data

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

class dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

class Smile_GAN_train():
    
    def __init__(self, ncluster, start_saving_epoch, max_epoch, WD_threshold, AQ_threshold, \
        cluster_loss_threshold, lam=9, mu=5, batchSize=25, lipschitz_k = 0.5, \
        beta1 = 0.5, lr = 0.0002, max_gnorm = 100, eval_freq = 25, save_epoch_freq = 5, print_freq = 1000):
        self.opt=dotdict({})
        self.opt.ncluster = ncluster
        self.opt.start_saving_epoch = start_saving_epoch
        self.opt.max_epoch = max_epoch
        self.opt.WD_threshold = WD_threshold
        self.opt.AQ_threshold = AQ_threshold
        self.opt.cluster_loss_threshold = cluster_loss_threshold
        self.opt.lam = lam
        self.opt.mu = mu
        self.opt.batchsize = batchSize
        self.opt.lipschitz_k = lipschitz_k
        self.opt.beta1 = beta1
        self.opt.lr = lr
        self.opt.max_gnorm = max_gnorm
        self.opt.save_epoch_freq = save_epoch_freq
        self.opt.print_freq = print_freq
        self.opt.eval_freq = eval_freq

    def print_log(self, result_f, message):
        result_f.write(message+"\n")
        result_f.flush()
        print(message)

    def format_log(self, epoch, epoch_iter, measures, t, prefix=True):
        message = '(epoch: %d, iters: %d, time: %.3f) ' % (epoch, epoch_iter, t)
        if not prefix:
            message = ' ' * len(message)
        for key, value in measures.items():
            message += '%s: %.4f ' % (key, value)
        return message

    def parse_data(self, data, covariate, random_seed, data_fraction):
        cn_train_dataset, pt_train_dataset, correction_variables, normalization_variables = parse_train_data(data, covariate, random_seed, data_fraction, self.opt.batchsize)
        cn_eval_dataset, pt_eval_dataset = parse_validation_data(data, covariate, correction_variables, normalization_variables)
        self.opt.nROI = pt_eval_dataset.shape[1]
        self.opt.n_val_data = pt_eval_dataset.shape[0]
        return cn_train_dataset, pt_train_dataset, cn_eval_dataset, pt_eval_dataset, correction_variables, normalization_variables


    def train(self, model_name, data, covariate, save_dir, random_seed=0, data_fraction=1, verbose=True, independent_ROI = True):
        if verbose: result_f = open("%s/results.txt" % save_dir, 'w')

        cn_train_dataset, pt_train_dataset, eval_X, eval_Y, correction_variables, normalization_variables = self.parse_data(data, covariate, random_seed, data_fraction)
        self.opt.correction_variables = correction_variables
        self.opt.normalization_variables = normalization_variables
        
        # create_model
        model = SmileGAN()
        model.create(self.opt)

        total_steps = 0
        print_start_time = time.time()
        max_distance_list = [0,0]                    ##### number of consecutive epochs with max_wd < threshold
        aq_loss_cluster_list = [[0 for _ in range(4)] for _ in range(2)]                    ##### number of consecutive epochs with aq and cluster_loss < threshold
        early_retrain_list = [0 for _ in range(10)]

        savetime=0                                      ##### keep track of number of times that stopping criteria is satisfied if choosing not to stop model
        predicted_label_past=np.zeros(self.opt.n_val_data)   ##### keep track of assigned clustering membership in epochs
        if not verbose:
            pbar = tqdm(total = self.opt.max_epoch)
        for epoch in range(1, self.opt.max_epoch + 1):
            if not verbose:
                pbar.update(1)
            epoch_start_time = time.time()
            epoch_iter = 0
            for i, pt_data in enumerate(pt_train_dataset):
                cn_data = cn_train_dataset.next()
                real_X, real_Y = Variable(cn_data['x']), Variable(pt_data['y'])

                total_steps += self.opt.batchsize
                epoch_iter += self.opt.batchsize

                losses = model.train_instance(real_X, real_Y)

                if total_steps % self.opt.print_freq == 0:
                    t = (time.time() - print_start_time) / self.opt.batchsize
                    if verbose:
                        self.print_log(result_f, self.format_log(epoch, epoch_iter, losses, t))
                        print_start_time = time.time()

            if epoch % self.opt.save_epoch_freq == 0 and verbose:
                self.print_log(result_f, 'saving the model at the end of epoch %d.' % (epoch))
                model.save(save_dir,'latest')

            if epoch % self.opt.eval_freq == 0:
                t = time.time()
            
                predicted_label,predicted_class = label_change(model,eval_Y,self.opt)
                total_label_change = np.count_nonzero(np.absolute(np.array(predicted_label)-np.array(predicted_label_past))) 
                predicted_label_past = predicted_label
                max_distance, w_distances = eval_w_distances(eval_X, eval_Y, model, independent_ROI)
                max_distance_list.append(max_distance)
                max_distance_list.pop(0)
                aq_loss_cluster_list[0].append(total_label_change)
                aq_loss_cluster_list[1].append(losses['loss_cluster'])
                early_retrain_list.append(losses['loss_cluster'])
                for _ in range(2):
                    aq_loss_cluster_list[_].pop(0) 
                early_retrain_list.pop(0)

            
                t = time.time() - t
                res_str_list = ["[%d], Mean_W_Distance: %.4f, TIME: %.4f" % (epoch,max_distance, t)]
                res_str_list.extend(['W_Distances: %s' % [round(ele, 4) for ele in w_distances],'Subtype_Quantity: %s' %predicted_class])

                if min(early_retrain_list)>0.4:
                    return False

                if len(w_distances)==model.opt.ncluster:
                    if max(max_distance_list) < self.opt.WD_threshold and max(aq_loss_cluster_list[0]) < self.opt.AQ_threshold\
                     and max(aq_loss_cluster_list[1])<self.opt.cluster_loss_threshold and epoch > self.opt.start_saving_epoch:
                        savetime+=1
                        model.save(save_dir, model_name)
                        if not verbose: pbar.close()
                        res_str_list += ["*** Stopping Criterion Satisfied ***"]
                        res_str = "\n".join(["-"*60] + res_str_list + ["-"*60])
                        if verbose: 
                            self.print_log(result_f, res_str)
                            result_f.close()
                        return True

                res_str = "\n".join(["-"*60] + res_str_list + ["-"*60])
                res_str_list += ["*** Max iteration reached and criterion not satisfied ***"]
                if verbose:  
                    self.print_log(result_f, res_str)
        if verbose: result_f.close()
        if not verbose: pbar.close()
        return False



import numpy as np
import scipy
import pandas as pd
import itertools
from sklearn.preprocessing import minmax_scale
from sklearn.linear_model import LinearRegression
from sklearn.cluster import KMeans
from sklearn import metrics
from .data_loading import PTIterator, CNIterator, val_PT_construction, val_CN_construction

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

def find_highest_ari_model(clustering_results):
    """
    Find one of save models which have the highest average overlap (evaluated by ARI) with all other saved models
    The returned model will be used as a template such that all other models will \
    be reordered to achieve the highest overlap with it.
    :param clustering_results: list, list of clustering results given by all saved models; 
                               length of list equals the number of saved models.
    :return: int, the index of selected model with highest average overlap
    """
    highest_ari=0
    best_model=0
    for i in range(len(clustering_results)):
        all_ari=[]
        for j in range(len(clustering_results)):
            if i!=j:all_ari.append(metrics.adjusted_rand_score(clustering_results[i], clustering_results[j]))
        if np.mean(all_ari)>highest_ari:
            best_model=i
            highest_ari=np.mean(all_ari)
    return best_model


def get_model_order(cluster_results,ncluster):
    """
    Find best orders for results given by all saved models so that they reach highest agreements
    :param clustering_results: list, list of clustering results given by all saved models; 
                               length of list equals the number of saved models.
    :param ncluster: int, number of clusters
    :return: list, list of best orders for results given by all saved models.
    """
    order_permutation = list(itertools.permutations(range(ncluster)))
    best_model = find_highest_ari_model(cluster_results)
    all_orders=[]
    for k in range(len(cluster_results)):
        if k==best_model: 
            all_orders.append(range(ncluster))
        elif k!=best_model:
            highest_intersection = 0
            for order in order_permutation:
                total_intersection = 0
                for i in range(ncluster):
                    total_intersection+= np.intersect1d(np.where(cluster_results[best_model]==i),np.where(cluster_results[k]==order[i])).shape[0]
                if total_intersection>=highest_intersection:
                    best_order=order
                    highest_intersection=total_intersection
            all_orders.append(best_order)
    return all_orders

def highest_matching_clustering(clustering_results, label_probability, ncluster):
    """
    The function which offers clustering result (cluster label and cluster probabilities) 
    by reordering and combining clustering results given by all saved models
    :param clustering_results: list, list of clustering results given by all saved models; 
                               length of list equals the number of saved models.
    :param label_probability: list, list of clutering results given by all saved modesl;
                                length of list equals the number of saved models and each 
                                model gives an n*k list.
    :param ncluster: int, number of clusters
    :return: two arrays, one n*1 array with cluster label for each participant
                         one n*k array with k cluster probabilites for each participant
    """
    order = get_model_order(clustering_results, ncluster)
    class_index=0
    for i in range(len(clustering_results)):
        label_probability[i] = label_probability[i][:,order[i]]
    prediction_prob=np.mean(label_probability,axis=0)
    prediction_cluster=prediction_prob.argmax(axis=1)
    return prediction_cluster, prediction_prob, 

def Covariate_correction(cn_data,cn_cov,pt_data,pt_cov):
    """
    Eliminate the confound of covariate, such as age and sex, from the disease-based changes.
    :param cn_data: array, control data
    :param cn_cov: array, control covariates
    :param pt_data: array, patient data
    :param pt_cov: array, patient covariates
    :return: corrected control data & corrected patient data
    """
    min_cov = np.amin(cn_cov, axis=0)
    max_cov = np.amax(cn_cov, axis=0)
    pt_cov = (pt_cov-min_cov)/(max_cov-min_cov)
    cn_cov = (cn_cov-min_cov)/(max_cov-min_cov)
    beta = np.transpose(LinearRegression().fit(cn_cov, cn_data).coef_)
    corrected_cn_data = cn_data-np.dot(cn_cov,beta)
    corrected_pt_data = pt_data-np.dot(pt_cov,beta)
    correction_variables = {'max_cov':max_cov,'min_cov':min_cov,'beta':beta}
    return corrected_cn_data, corrected_pt_data, correction_variables

def Data_normalization(cn_data,pt_data):
    """
    Normalize all data with respect to control data to ensure a mean of 1 and std of 0.1 
    among CN participants for each ROI
    :param cn_data: array, control data
    :param pt_data: array, patient data
    :return: normalized control data & normalized patient data
    """
    cn_mean = np.mean(cn_data,axis=0)
    cn_std = np.std(cn_data,axis=0)
    normalized_cn_data = 1+(cn_data-cn_mean)/(10*cn_std)
    normalized_pt_data = 1+(pt_data-cn_mean)/(10*cn_std)
    normalization_variables = {'cn_mean':cn_mean, 'cn_std':cn_std}
    return normalized_cn_data, normalized_pt_data, normalization_variables

def apply_covariate_correction(data, covariate, correction_variables):
    covariate = (covariate-correction_variables['min_cov'])/(correction_variables['max_cov']-correction_variables['min_cov'])
    corrected_data = data-np.dot(covariate,correction_variables['beta'])
    return corrected_data

def apply_data_normalization(data, normalization_variables):
    normalized_data = 1+(data-normalization_variables['cn_mean'])/(10*normalization_variables['cn_std'])
    return normalized_data

def parse_train_data(data, covariate, random_seed, data_fraction, batch_size):
    cn_data = data.loc[data['diagnosis'] == -1].drop(['participant_id','diagnosis'], axis=1).values
    pt_data = data.loc[data['diagnosis'] == 1].drop(['participant_id','diagnosis'], axis=1).values
    correction_variables = None
    if covariate is not None:
        cn_cov = covariate.loc[covariate['diagnosis'] == -1].drop(['participant_id', 'diagnosis'], axis=1).values
        pt_cov = covariate.loc[covariate['diagnosis'] == 1].drop(['participant_id','diagnosis'], axis=1).values
        cn_data,pt_data, correction_variables = Covariate_correction(cn_data,cn_cov,pt_data,pt_cov)
    normalized_cn_data, normalized_pt_data, normalization_variables = Data_normalization(cn_data,pt_data)
    cn_train_dataset = CNIterator(normalized_cn_data, random_seed, data_fraction, batch_size)
    pt_train_dataset = PTIterator(normalized_pt_data, random_seed, data_fraction, batch_size)
    return cn_train_dataset, pt_train_dataset, correction_variables, normalization_variables


def parse_validation_data(data, covariate, correction_variables, normalization_variables):
    cn_data = data.loc[data['diagnosis'] == -1].drop(['participant_id','diagnosis'], axis=1).values
    pt_data = data.loc[data['diagnosis'] == 1].drop(['participant_id','diagnosis'], axis=1).values
    if correction_variables is not None:
        cn_cov = covariate.loc[covariate['diagnosis'] == -1].drop(['participant_id', 'diagnosis'], axis=1).values
        pt_cov = covariate.loc[covariate['diagnosis'] == 1].drop(['participant_id','diagnosis'], axis=1).values
        cn_data = apply_covariate_correction(cn_data, cn_cov, correction_variables)
        pt_data = apply_covariate_correction(pt_data, pt_cov, correction_variables)
    normalized_cn_data = apply_data_normalization(cn_data, normalization_variables)
    normalized_pt_data = apply_data_normalization(pt_data, normalization_variables)
    cn_eval_dataset = val_CN_construction(normalized_cn_data).load()
    pt_eval_dataset = val_PT_construction(normalized_pt_data).load()
    return cn_eval_dataset, pt_eval_dataset



def check_symmetric(a, rtol=1e-05, atol=1e-08):
    """
    Check if the numpy array is symmetric or not
    """
    result = np.allclose(a, a.T, rtol=rtol, atol=atol)
    return result

def consensus_clustering(clustering_results, k):
    """
    This function performs consensus clustering on a co-occurence matrix
    :param clustering_results: an array containing all the clustering results given by all saved models
    :param k: number of clusters
    :return:
    """
    clustering_results = np.transpose(np.array(clustering_results))
    num_pt = clustering_results.shape[0]
    cooccurence_matrix = np.zeros((num_pt, num_pt))

    for i in range(num_pt - 1):
        for j in range(i + 1, num_pt):
            cooccurence_matrix[i, j] = sum(clustering_results[i, :] == clustering_results[j, :])

    cooccurence_matrix = np.add(cooccurence_matrix, cooccurence_matrix.transpose())
    ## here is to compute the Laplacian matrix
    Laplacian = np.subtract(np.diag(np.sum(cooccurence_matrix, axis=1)), cooccurence_matrix)

    Laplacian_norm = np.subtract(np.eye(num_pt), np.matmul(np.matmul(np.diag(1 / np.sqrt(np.sum(cooccurence_matrix, axis=1))), cooccurence_matrix), np.diag(1 / np.sqrt(np.sum(cooccurence_matrix, axis=1)))))
    ## replace the nan with 0
    Laplacian_norm = np.nan_to_num(Laplacian_norm)

    ## check if the Laplacian norm is symmetric or not, because matlab eig function will automatically check this, but not in numpy or scipy
    if check_symmetric(Laplacian_norm):
        ## extract the eigen value and vector
        ## matlab eig equivalence is eigh, not eig from numpy or scipy, see this post: https://stackoverflow.com/questions/8765310/scipy-linalg-eig-return-complex-eigenvalues-for-covariance-matrix
        ## Note, the eigenvector is not unique, thus the matlab and python eigenvector may be different, but this will not affect the results.
        evalue, evector = scipy.linalg.eigh(Laplacian_norm)
    else:
        # evalue, evector = np.linalg.eig(Laplacian_norm)
        raise Exception("The Laplacian matrix should be symmetric here...")

    ## check if the eigen vector is complex
    if np.any(np.iscomplex(evector)):
        evalue, evector = scipy.linalg.eigh(Laplacian)

    ## create the kmean algorithm with sklearn
    kmeans = KMeans(n_clusters=k, n_init=20).fit(evector.real[:, 0: k])
    final_predict = kmeans.labels_

    return final_predict






import numpy as np
from scipy.linalg import sqrtm
from .model import sample_z

__author__ = "Zhijian Yang"
__copyright__ = "Copyright 2019-2020 The CBICA & SBIA Lab"
__credits__ = ["Zhijian Yang"]
__license__ = "See LICENSE file"
__version__ = "0.1.0"
__maintainer__ = "Zhijian Yang"
__email__ = "zhijianyang@outlook.com"
__status__ = "Development"

## calulate wasserstein distance assuming independent ROIs
def cal_w_distance(mean_1,mean_2,std_1,std_2):
    distance=0
    for i in range(mean_1.size):
        distance+=((mean_1[i]-mean_2[i])**2+(std_1[i])**2+std_2[i]**2-2*std_1[i]*std_2[i])
    return  distance
    
## calulate wasserstein distance without assuming independent ROIs
def cal_w_distance_with_cov(mean_1,mean_2,std_1,std_2):
    distance=0
    distance=np.sum(np.square(mean_1-mean_2))+np.trace(std_1+std_2-2*sqrtm(sqrtm(std_2).dot(std_1).dot(sqrtm(std_2))))
    return  distance.real

## return mean wasserstein distance and wasserstein distance of each mapping direction
def cal_validate_distance(model,mean,std,eva_data,independent):
    predicted_z=model.predict_cluster(eva_data)
    cluster=[[]for k in range(model.opt.ncluster)]
    distance=[1000 for m in range(model.opt.ncluster)]
    adjusted_eva_data=eva_data.detach().numpy()
    for i in range(predicted_z.shape[0]):
        a=predicted_z[i].tolist().index(max(predicted_z[i].tolist()))
        cluster[a].append(adjusted_eva_data[i])
    for j in range(model.opt.ncluster):
        if len(cluster[j])>1:
            cluster_mean=np.mean(np.array(cluster[j]), axis=0)
            if independent:
                cluster_std=np.std(np.array(cluster[j]), axis=0)
                distance[j]=cal_w_distance(mean[j],cluster_mean,std[j],cluster_std)
            else:
                cluster_std=np.cov(np.transpose(np.array(cluster[j])))
                distance[j]=cal_w_distance_with_cov(mean[j],cluster_mean,std[j],cluster_std)
    output=[0 for _ in range(model.opt.ncluster)]
    for k in range(model.opt.ncluster):
        if distance[k]!=1000:
            output[k] = distance[k]
    return np.mean(output),output
      
def eval_w_distances(real_X,real_Y,model,independent):
    predict_Y=[]
    mean=[]
    std=[]
    for i in range(model.opt.ncluster):
        z, z_index = sample_z(real_X, model.opt.ncluster, fix_class=i)
        predict_Y.append(model.predict_Y(real_X,z).detach().numpy())
    for j in range(model.opt.ncluster):
        mean.append(np.mean(predict_Y[j], axis=0))
        std.append(np.std(predict_Y[j], axis=0)) if independent else std.append(np.cov(np.transpose(np.array(predict_Y[j]))))
    max_distance, w_distances=cal_validate_distance(model,mean,std,real_Y,independent)
    return max_distance, w_distances

## return clustering memberships of valY datapoints and quantity of subjects assigned to each subtype
def label_change(model,test_Y,opt):
    predicted_label=[]
    predicted_class=[0 for _ in range(opt.ncluster)]
    prediction=model.predict_cluster(test_Y)
    for i in range(prediction.shape[0]):
        label=prediction[i].tolist().index(max(prediction[i].tolist()))
        predicted_label.append(label)
        predicted_class[label]+=1
    return predicted_label, predicted_class






